"""Parameter enums"""


class ParameterTypes:
    """Standard parameter types"""
    STR: str = 'str'
    INT: str = 'int'
    FLOAT: str = 'float'
    BOOL: str = 'bool'
    TEXT: str = 'text'
    ENUM: str = 'enum'
    LIST_STR: str = 'list_str'
    LIST_INT: str = 'list_int'
    LIST_NODE: str = 'list_node'
    CODE: str = 'code'
